from Foundation import *
from AppKit import *
import objc

# Local imports from within Circuit Editor
from CE.Wire import Wire, WireEnd
from geometry import *
from pointmap import pointmap
from notifications import *

import circuitsim
from solver4 import SolverError

class Circuit(NSObject):
	"""A Circuit represents a circuit board. The parts on the circuit board are wires and 
components."""
	
	# A Circuit object has the following fields:
	
	# circuit: A circuitsim.Circuit object.
	
	# width: An integer.
	# height: An integer.
	
	# wires: A set of CE.Wire objects.
	# components: A set of CE.Component objects.
	# nodes: A set of CE.Node objects.
	
	# "map" and "secondary_map" speed up the finding of connections between elements of the
	# circuit. Each one maps (x,y) coordinate pairs to sets of elements. Each component is
	# in "map" at the locations of its pins, and each wire is in "map" at the locations of
	# its endpoints. In addition, each wire is in "secondary_map" at any point where it
	# passes through a lattice point.
	
	# map: A pointmap.pointmap object of components and wires
	# secondary_map: A pointmap.pointmap object of components and wires
	
	# If the circuit's calculation was updated recently and the circuit has not changed since
	# then, then "calculation_valid" is True. Otherwise, it is False.
	
	# calculation_valid: A Boolean value.
	
	def initEmpty(self):
		"""Initializes the circuit board with a grid of 30x30 squares and no components."""
		
		self.circuit = circuitsim.Circuit()
			
		# TODO: make it possible to have boards of different sizes
		self.width = 30
		self.height = 30
		
		self.wires = set()
		self.components = set()
		
		self.nodes = set()
		
		self.map = pointmap()
		self.secondary_map = pointmap()
		
		self.calculation_valid = False
		
		self.setUndoFunc_(None)
		
		return self
	
	def initWithElements_(self, objs):
		"""Initializes the circuit board with a set of wires and components, presumably loaded from
		a file."""
		
		self.initEmpty()
		
		for obj in objs:
			# WTF: Why don't we just use obj.install() here instead?
			obj.circuit = self
			obj.install_in_map()
			if isinstance(obj, Wire):
				self.wires.add(obj)
			else:
				self.components.add(obj)
				obj.component.install(self.circuit)
		
		return self
	
	def setUndoFunc_(self, func):
		"""Sets the function to be called when the circuit wants to register an undo action."""
		if func:
			self.undoer = func
		else:
			self.undoer = lambda f:None
	
	def recalculate(self):
		"""recalculate() tells the solver to try to determine voltage and current values everywhere 
		in the circuit."""
		
		# Eliminate unconnected node objects so the solver doesn't try to find values for their
		# voltages
		nodes_to_destroy = []
		for node in self.nodes:
			if not node.users:
				nodes_to_destroy.append(node)
		for node in nodes_to_destroy:
			node.destroy()
		
		try:
			self.circuit.calculate()
		except SolverError, e:
			self.calculation_valid = False
			post_notification("CircuitRecalculated", self)
			raise e # CircuitDocument will catch it and display an alert
		else:
			self.calculation_valid = True
			post_notification("CircuitRecalculated", self)
	
	def update_(self, t):
		"""update_(t) tells the solver to advance time in Circuit-Land by 't' seconds."""
		
		if not self.calculation_valid:
			self.recalculate()
		
		try:
			self.circuit.calculate_change(t)
		except SolverError, e:
			self.calculation_valid = False
			post_notification("CircuitRecalculated", self)
			raise e
		else:
			self.calculation_valid = True
			post_notification("CircuitRecalculated", self)