from __future__ import with_statement

from Foundation import *
from AppKit import *
import objc

# Local imports from within Circuit Editor
from geometry import *
from draw_extras import *
from notifications import post_notification
from CE.Node import Node

import circuitsim

import traceback

# Text attribute dictionary for voltage labels
voltage_label_attribs = {
	NSFontAttributeName: load_font("Helvetica", 10),
	NSBackgroundColorAttributeName: NSColor.darkGrayColor(),
	NSForegroundColorAttributeName: NSColor.whiteColor()
	}

class Component(object):
	
	"""A CE.Component object describes a single component in a circuit. Component itself is an
abstract class; use its concrete subclasses such as CE.Components.Resistor, CE.Components.Capacitor, 
etc. instead.

Components can exist independently outside of a circuit until they are installed in a circuit."""
	
	
	
	# The subclass of Component define the following class members:
	
	# width, height: The size of an instance of the component. Mandatory.
	
	# pin_locations: A list of coordinates, one point for each pin. Mandatory.
	
	# class_name: A human-readable name for the component class. Mandatory.
	
	# editable_properties: A list of (name, ui_name, type) which specifies what properties the
	#	component has that can be edited using the user interface. If the subclass does not specify
	#   editable_properties, then it will default to the definition below.
	
	# The default 'name' property is to help make debug messages more informative
	editable_properties = [("name", "Name", "string")]
	
	# dynamic_appearance: Boolean value. True if the component's appearance depends on the amount
	#   of current going through it or on its voltage, False otherwise. Optional.
	dynamic_appearance = False

	# manipulation_zones: A list of rectangles in the component's coordinate system, indicating the
	#   areas where the component responds to the mouse. The component will only receive
	#   manipulation information if the mouse starts in one of its manipulation zones. Optional.
	manipulation_zones = []
	
	
	
	# A Component instance has the following fields:
	
	# center: A point, each element of which is a multiple of 1/2.
	# rotation: A number in degrees.
	# flip: A boolean value.
	
	# Together, "center", "rotation", and "flip" define the location and orientation of the
	# component.
	
	# transform: A NSAffineTransform object.
	
	# "transform" holds a transformation between the circuit's coordinate system and the component's
	# coordinate system. It is determined by "center", "rotation", and "flip"
	
	
	
	def __init__(self, serialized_form = None):
		self.circuit = None
		
		if serialized_form:
			# Unpack the serialized version if there is one
			self.center, self.rotation, self.flip, name = serialized_form
			if name: self.component.name = name
			
		else:
			# Default to put lower-left corner at (0,0)
			self.center = (self.width/2.0, self.height/2.0)
			self.rotation = 0
			self.flip = False
		
		# Make sure the component is aligned with the grid
		if self.rotation%180 == 0: w,h = self.width, self.height
		else: w,h = self.height, self.width
		assert (self.center[0]+w/2.0) % 1 == 0
		assert (self.center[1]+h/2.0) % 1 == 0
		
		self.transform = make_affine_transform(self.center, self.rotation, self.flip)
		
		self.pins = []
		for pin_id in xrange(len(self.pin_locations)):
			self.pins.append(ComponentPin(self, pin_id))
	
	def install_in_map(self):
		"""Do not call this directly.
		
install_in_map() helps maintain the circuit's map of what is installed where, as well as keeping 
track of what node each pin is connected to. It is called whenever the component is installed in a 
new circuit or moves to a new location within a circuit."""
		
		for pin in self.pins:
			self.circuit.map.add(pin.location(), pin)
		
		for pin in self.pins:
			pin.set_node(Node(self.circuit))
	
	def remove_from_map(self):
		"""Do not call this directly.
		
remove_from_map() helps maintain the circuit's map of what is installed where, in addition to 
keeping track of what node each pin and wire is connected to. It is called whenever the component is 
removed from a circuit or moves to a new location within a circuit."""
		
		for pin in self.pins:
			self.circuit.map.remove(pin.location(), pin)
		
		for pin in self.pins:
			pin.set_node(None)
			
			if not self.circuit.map.anything_at(pin.location()):
				# It's possible that the pin was "holding together" two wires that would otherwise # # cross without touching. In this case, assign each wire a different node.

				for connection in self.circuit.secondary_map.get(pin.location()):
					connection.set_node(Node(self.circuit))
	
	def install(self, circuit):
		"""install() installs the component in the given circuit."""
		assert not self.circuit
		
		self.circuit = circuit
		circuit.components.add(self)
		self.component.install(circuit.circuit)
		self.install_in_map()
		
		post_notification("CircuitElementAdded", circuit, self)
		circuit.calculation_valid = False
				
		circuit.undoer(lambda:self.remove())
	
	def reposition(self, new_position):
		"""Changes the position of the component within the circuit. If the component is not in a
circuit, then change 'center', 'rotation', and 'flip' directly, and then rebuild 'transform'
manually."""
		assert self.circuit
		
		# Make sure the component is aligned with the grid
		if new_position[1]%180 == 0: w,h = self.width, self.height
		else: w,h = self.height, self.width
		assert (new_position[0][0]+w/2.0) % 1 == 0
		assert (new_position[0][1]+h/2.0) % 1 == 0
		
		self.remove_from_map()
		
		old_position = (self.center, self.rotation, self.flip)
		self.center, self.rotation, self.flip = new_position
		self.transform = make_affine_transform(*new_position)
		
		self.install_in_map()
		
		post_notification("CircuitElementRepositioned", self.circuit, self)
		self.circuit.calculation_valid = False
		self.circuit.undoer(lambda:self.reposition(old_position))
	
	def remove(self):
		"""Removes the component from the circuit it is in."""
		assert self.circuit
		circuit = self.circuit
		
		post_notification("CircuitElementRemoved", circuit, self)
		circuit.calculation_valid = False
		
		self.remove_from_map()
		self.component.remove()
		circuit.components.remove(self)
		self.circuit = None
		
		circuit.undoer(lambda:self.install(circuit))
	
	def get_property(self, property):
		"""Asks for a named attribute of the component. The property names come from the
'editable_properties' attribute of the component. Subclasses which define more editable properties
may override this function, but do not have to; if the property corresponds to an attribute of the
same name in the Component object, then it will be found automatically."""
		if property=="name":
			if self.component.name:
				return self.component.name
			else:
				return ""
		else:
			return getattr(self, property)
	
	def set_property(self, property, new_value):
		"""Changes a named attribute of the component. The property names come from the
'editable_properties' attribute of the component. Subclasses which defined more editable properties
should override this function."""
		if property=="name":
			self.component.name = new_value
		else:
			raise Exception("Don't know about any property called '%s'" % property)
	
	def change_property(self, property, new_value):
		"""change_property is like set_property except that it registers an undo callback and posts
a notification in addition to changing the property."""
		
		old_value = self.get_property(property)
		
		self.set_property(property, new_value)
		
		post_notification("CircuitElementPropertyChange", self.circuit, (self, property))
		self.circuit.undoer(lambda:self.change_property(property, old_value))
	
	def draw(self):
		"""draw() is called by the view object when it is time for the component to draw itself. The
coordinate system when draw() is called will be the circuit's coordinate system."""
		
		with in_affine_transform(self.transform):
			self.subdraw()
		
		NSColor.blackColor().set()
		for pin in self.pins:
			r = rect_around_point(transform_point(self.transform, pin.local_location), 0.2)
			NSBezierPath.bezierPathWithOvalInRect_(r).fill()
		
	def subdraw(self):
		"""subdraw() is called by the component object itself when the component needs to be drawn. 
The current graphics transformation will be the appropriate one to move and orient the component, so 
subdraw() should just draw the component as though it is at the origin and oriented normally."""
		raise NotImplementedError
	
	def draw_outline(self, transform=None, color=NSColor.greenColor()):
		"""draw_outline() is called by the view object when the component is soon to be dropped. The
outline is used to show where the component will go when it is dropped. The coordinate system in use 
will be the circuit's coordinate system."""
		
		b = self.bounds()
		
		if transform and self in transform[1]:
			b = transform_rect(transform[0], b)
		
		color.set()
		NSBezierPath.strokeRect_(b)
	
	def draw_labels(self):
		"""draw_labels() is called by the view object when the component should draw voltage
labels"""
		
		if self.circuit and self.circuit.calculation_valid:
			# If the calculation is invalid, then the voltage values will be meaningless.
		
			for pin in self.pins:
			
				voltage = pin.node.voltage()
				if voltage>0:
					text = "+%.1f" % voltage
				elif voltage<0:
					text = "-%.1f" % abs(voltage)
				else:
					text = "0.0"
				
				loc = pin.location()
				loc = (loc[0], loc[1]+0.5)
							
				draw_text_centered(text, loc, voltage_label_attribs)
	
	def draw_selection(self, transform=None):
		"""draw_selection() is called by the view object to display a selection box around the
component."""
		b = self.bounds()
		if transform and self in transform[1]:
			b = transform_rect(transform[0], b)
		
		NSColor.blueColor().set()
		NSBezierPath.strokeRect_(expand_rect(b, 0.2))
	
	def start_manipulation(self, point):
		"""start_manipulation() is called by the view object if the component is clicked on in one
of the manipulatable zones designated by 'manipulation_zones'. The argument is the location of the
mouse in component coordinates."""
		pass
	
	def continue_manipulation(self, point):
		"""continue_manipulation() is called by the view object if the mouse is dragged after
start_manipulation()."""
		pass
	
	def end_manipulation(self, point):
		"""end_manipulation() is called by the view object when the mouse is released after
start_manipulation()."""
		pass
	
	def transformed_position(self, func):
		"""Returns what the component's "center", "rotation", and "flip" values would be after it
was transformed by 'func', which is an NSAffineTransform object."""
		func2 = self.transform.copy()
		func2.appendTransform_(func)
		return disect_affine_transform(func2)

	def bounds(self):
		"""bounds() returns a bounding rectangle for the component in the circuit's coordinate
system."""
		local_bounds = ((-self.width/2.0, -self.height/2.0), (self.width, self.height))
		return transform_rect(self.transform, local_bounds)
	
	def serialize(self):
		"""serialize() converts the component into a form suitable for converting into a string. The
return value of serialize() does not have to be a string, but it must be an object such that 
eval(repr(object)) == object."""
		return (self.center, self.rotation, self.flip, self.component.name)

class ComponentPin(object):

	"""A ComponentPin object represents one pin of a single component object."""
	
	def __init__(self, component, pin_id):
		# The component object that this pin is a part of
		self.component = component
		
		# The number of the pin on the component object
		self.pin_id = pin_id
		
		# The node that this pin is currently attached to
		self.node = None
		
		# The location of the pin in the component's coordinate system
		self.local_location = self.component.pin_locations[pin_id]
	
	def location(self):
		# Returns the location of the pin in the circuit's coordinate system
		return round_point(transform_point(self.component.transform, self.local_location))
	
	def set_node(self, node):
		# Disconnects the pin from any node it is connected to and reconnects it to the given node.
		# Also, propagates the given node to any wires or pins this pin is connected to.
		
		if self.node != node:
			
			# If we already were connected to a node, disconnect our internal circuitsim.Component
			# from it, and update node users list
			if self.node:
				circuitsim.unwire(self.component.component, self.pin_id)
				self.node.users.remove(self)
			
			self.node = node
			
			# Propagate node to connected wires and pins
			circuit = self.component.circuit
			if node:
				for connection in circuit.map.get(self.location()):
					connection.set_node(node)
				for connection in circuit.secondary_map.get(self.location()):
					connection.set_node(node)
			
			# Connect node to internal circuitsim.Component object, and update node users list
			if self.node:
				circuitsim.wire(self.component.component, self.pin_id, node.node)
				self.node.users.add(self)
				