from Foundation import *
from AppKit import *
import objc

# Local imports from within Circuit Editor
from geometry import *
from notifications import post_notification
from CE.Node import *

# Wire objects can exist independent of a Circuit object.

class Wire(object):
	def __init__(self, (start, end)):
		# The circuit that the wire is a part of (initially the wire is "orphaned")
		self.circuit = None
		
		# The two ends of the wire
		self.start = WireEnd(self, start)
		self.end = WireEnd(self, end)
		
		# The node that the wire is currently attached to
		self.node = None
	
	def install_in_map(self):
		# install_in_map() helps maintain the circuit's map of what is installed where, as well as
		# keeping track of what node each wire and pin is connected to. It is called whenever the 
		# wire is installed in a new circuit or moves to a new location within a circuit.
		
		self.circuit.map.add(self.start.location, self)
		if self.start.location != self.end.location:
			self.circuit.map.add(self.end.location, self)
		for p in iter_lattice_points(self.start.location, self.end.location):
			self.circuit.secondary_map.add(p, self)
		
		self.set_node(Node(self.circuit))
	
	def remove_from_map(self):
		# install_in_map() helps maintain the circuit's map of what is installed where, as well as
		# keeping track of what node each wire and pin is connected to. It is called whenever the 
		# wire is removed from a circuit or moves to a new location within a circuit.
		
		self.circuit.map.remove(self.start.location, self)
		if self.start.location != self.end.location:
			self.circuit.map.remove(self.end.location, self)
		for p in iter_lattice_points(self.start.location, self.end.location):
			self.circuit.secondary_map.remove(p, self)
		
		# node_generator generates a supply of new nodes. The first node generated will be the
		# original node of the wire, so that it can be re-used if possible. Subsequent nodes will
		# be generated as necessary.
		def node_generator():
			yield self.node
			while True:
				yield Node(self.circuit)
		generator = node_generator()
		
		# Assign a new node to each pin or wire that is on the wire end
		for point in set([self.start.location, self.end.location]):
			# If there are one or more pins or wire ends at the ends of this wire, then simply
			# assign the first one a new node, and it will propagate that node to any other pins,
			# wire-ends, or wire-midpoints. If there are no pins or wire ends, then check for
			# wire-midpoints, and assign each one a new node.
			if self.circuit.map.anything_at(point):
				for connection in self.circuit.map.get(point):
					connection.set_node(generator.next())
					break # Exit after the first one - it will propagate to the others on its own
			else:
				for connection in self.circuit.secondary_map.get(point):
					connection.set_node(generator.next())
		
		# Assign new node to each pin or wire end that is on the wire's body
		for point in iter_lattice_points(self.start.location, self.end.location):
			for connection in self.circuit.map.get(point):
				connection.set_node(generator.next())
				break # Exit after the first one - it will propagate to the others on its own
		
		self.set_node(None)
	
	def install(self, circuit):
		assert not self.circuit
		
		self.circuit = circuit
		circuit.wires.add(self)
		self.install_in_map()
		
		post_notification("CircuitElementAdded", circuit, self)	
		circuit.calculation_valid = False
		
		circuit.undoer(lambda:self.remove())
	
	def reposition(self, new_location):	
		assert self.circuit
		
		self.remove_from_map()
		
		old_location = (self.start.location, self.end.location)
		self.start.location, self.end.location = new_location
		
		self.install_in_map()
		
		post_notification("CircuitElementRepositioned", self.circuit, self)
		self.circuit.calculation_valid = False
		
		self.circuit.undoer(lambda:self.reposition(old_location))
	
	def remove(self):
		assert self.circuit
		circuit = self.circuit
		
		post_notification("CircuitElementRemoved", circuit, self)
		circuit.calculation_valid = False
		
		self.remove_from_map()
		circuit.wires.remove(self)
		self.circuit = None
		
		circuit.undoer(lambda:self.install(circuit))
	
	def draw(self, transform=None):
		# draw() draws the wire.
		#
		# If 'transform' is not None, then it should be a pair (affinetransform, objs) where
		# 'affinetransform' is a NSAffineTransform object and 'objs' is a set of objects being
		# transformed. The wire will be drawn with the given transform applied.
				
		if transform:
			start,end = self.transformed_location(transform[0], transform[1])
		else:
			start,end = self.start.location, self.end.location
		
		NSColor.blackColor().set()
		Wire.draw_wire(start, end)
	
	def draw_outline(self, transform=None, color=NSColor.greenColor()):
		# draw_outline() is called by the view object when the wire is soon to be dropped. The
		# outline is used to show where the wire will go when it is dropped. The coordinate
		# system in use will be the circuit's coordinate system.
		
		if transform:
			start,end = self.transformed_location(transform[0], transform[1])
		else:
			start,end = self.start.location, self.end.location
		
		color.set()
		Wire.draw_wire(start, end)
	
	def set_node(self, node):
		# set_node() changes the wire's node to the given node. It also automatically 'propagates'
		# that new node to any other wires and pins that the wire is attached to.
		if node!=self.node:
		
			if self.node:
				self.node.users.remove(self)
			self.node = node
			if self.node:
				self.node.users.add(self)
			
			if node:
				for connection in self.circuit.map.get(self.start.location):
					connection.set_node(node)
				for connection in self.circuit.secondary_map.get(self.start.location):
					connection.set_node(node)
				for connection in self.circuit.map.get(self.end.location):
					connection.set_node(node)
				for connection in self.circuit.secondary_map.get(self.end.location):
					connection.set_node(node)
				for point in iter_lattice_points(self.start.location, self.end.location):
					for connection in self.circuit.map.get(point):
						connection.set_node(node)
	
	def transformed_location(self, func, objs):
		# transformed_location returns a tuple of (startloc, endloc) which is where the wire will be
		# after the members of 'objs' are transformed by 'func'.
		start = self.start.location
		end = self.end.location
		if self.start in objs:
			start = round_point(transform_point(func, start))
		if self.end in objs:
			end = round_point(transform_point(func, end))
		return (start,end)
	
	def bounds(self):
		return rect_with_corners(self.start.location, self.end.location)
	
	@classmethod
	def draw_wire(cls, start, end):
		# draw_wire() draws a wire in the current Cocoa drawing color going from 'start' to 'end'.

		NSBezierPath.strokeLineFromPoint_toPoint_(start, end)
		
		h1 = WireEnd.calc_handle_location(start, end, 0.25)
		h2 = WireEnd.calc_handle_location(end, start, 0.25)
		
		handles = NSBezierPath.alloc().init()
		handles.setLineCapStyle_(NSRoundLineCapStyle)
		handles.setLineWidth_(0.25)
		handles.moveToPoint_(start)
		handles.lineToPoint_(h1)
		handles.moveToPoint_(h2)
		handles.lineToPoint_(end)
		handles.stroke()
	
	def serialize(self):
		# serialize() produces a serialized representation of the wire to put on the clipboard or
		# to save to a file. This method doesn't actually return a string; instead, it must return
		# an object such that eval(repr(object)) == object.
		return (self.start.location, self.end.location)

# A WireEnd object represents one *end* of a Wire object. The reason why the two ends are
# represented as separate objects is because it is possible for one end of a wire to be selected
# in the GUI without the other end being selected.
class WireEnd(object):
	def __init__(self, wire, location):
		self.location = round_point(location)
		self.wire = wire
	
	def draw_selection(self, transform=None):
		# Draws a blue selection highlight on this end of the wire
		
		start,end = self.location,self.other().location
		
		if transform:
			if self in transform[1]:
				start = transform_point(transform[0], start)
			if self.other() in transform[1]:
				end = transform_point(transform[0], end)
		
		NSColor.blueColor().set()
		WireEnd.draw_wire_selection(start, end)
	
	@classmethod
	def draw_wire_selection(cls, start, end):
		# Draws a selection on the 'start' end of a wire going from 'start' to 'end'.
		
		h = WireEnd.calc_handle_location(start, end, 0.25)
		selection = NSBezierPath.alloc().init()
		selection.setLineCapStyle_(NSRoundLineCapStyle)
		selection.setLineWidth_(0.35)
		selection.moveToPoint_(start)
		selection.lineToPoint_(h)
		selection.stroke()
	
	@classmethod
	def calc_handle_location(self, start, end, dist):
		# _calc_handle_location returns location of a point which is on the line between 'start' and
		# 'end', exactly 'dist' away from 'start'.
		if start==end:
			return start
		else:
			v = norm_vector(subtract_points(end, start))
			return (start[0]+v[0]*dist, start[1]+v[1]*dist)
	
	def other(self):
		# Returns the other end of the wire.
		if self==self.wire.start:
			return self.wire.end
		elif self==self.wire.end:
			return self.wire.start
		else:
			raise