from CE.Component import Component
from CE.Wire import Wire, WireEnd
from geometry import *

def test_drop(size, group1, group2, (func,parts)):
	# test_drop() checks if it is safe to place a group of components and wires in a
	# circuit. 'size' is the size of the circuit and 'group1' is a list of the components and
	# wires already in the circuit. 'group2' is the group of wires and components to be placed
	# after being transformed with the transformation (func,parts).
	
	wires1 = [w for w in group1 if isinstance(w,Wire)]
	components1 = [c for c in group1 if isinstance(c,Component)]
	
	failures = set()
	
	# Compute locations of all wires and components in group2
	wires2_and_locs = []
	components2_and_bounds = []
	for obj in group2:
		if isinstance(obj, Wire):
			new_loc = obj.transformed_location(func, parts)
			wires2_and_locs.append((obj, new_loc))
		else:
			components2_and_bounds.append((obj, transform_rect(func, obj.bounds())))
	
	# Check for collisions between group1 components and group2 components
	for c1 in components1:
		for (c2, c2bounds) in components2_and_bounds:
			if rects_intersect(c1.bounds(), c2bounds):
				failures.add(c2)
	
	# Check for group1 wires ending in a group2 component
	for w1 in wires1:
		for end in (w1.start.location, w1.end.location):
			for (c2, bounds) in components2_and_bounds:
				if point_in_rect(end, bounds):
					failures.add(c2)
	
	# Check for group2 wires ending in a group1 component
	for c1 in components1:
		bounds = c1.bounds()
		for (w2,w2pos) in wires2_and_locs:
			for end in w2pos:
				if point_in_rect(end, bounds):
					failures.add(w2)
	
	# Check for overlap between group1 wires and group2 wires
	for w1 in wires1:
		w1pos = (w1.start.location, w1.end.location)
		for (w2, w2pos) in wires2_and_locs:
			if line_segments_overlap(w1pos, w2pos):
				failures.add(w2)
	
	# Check for group2 wires overlapping other group2 wires (this is possible because the
	# transformation may only affect one end of the wire)
	for i, (a, a_pos) in enumerate(wires2_and_locs):
		for (b, b_pos) in wires2_and_locs[i+1:]:
			if line_segments_overlap(a_pos, b_pos):
				failures.add(a)
				failures.add(b)
	
	# Check for group2 wires outside of the circuit bounds
	for (w2, w2pos) in wires2_and_locs:
		for end in w2pos:
			if not point_in_rect(end, ((-1,-1),(size[0]+1,size[1]+1))):
				failures.add(w2)
	
	# Check for group2 components outside of the circuit bounds
	for (c2, c2bounds) in components2_and_bounds:
		for end in c2bounds:
			if not point_in_rect(c2bounds[0], ((-1,-1),(size[0]+1,size[1]+1))):
				failures.add(c2)
			if not point_in_rect(add_points(c2bounds[0], c2bounds[1]),
					((-1,-1),(size[0]+1,size[1]+1))):
				failures.add(c2)
	
	return failures

def test_transform(size, objs, (func, parts)):
	# test_transform checks if the parts in 'parts' can be transformed by 'func'. If they can,
	# it returns set(). If they cannot, it returns a set of all the wires and components that
	# are not transformable.
	
	failures = set()
	
	wires = set([o for o in objs if isinstance(o,Wire)])
	components = set([o for o in objs if isinstance(o,Component)])
	
	# Compute lists of new locations of all wires
	wires_transformed = []
	wires_nontransformed = []
	for wire in wires:
		if wire.start in parts or wire.end in parts:
			wires_transformed.append((wire, wire.transformed_location(func,parts)))
		else:
			wires_nontransformed.append((wire, (wire.start.location, wire.end.location)))
	
	# Compute lists of new boundary rects of all components
	components_transformed = []
	components_nontransformed = []
	for c in components:
		if c in parts:
			components_transformed.append((c, transform_rect(func, c.bounds())))
		else:
			components_nontransformed.append((c, c.bounds()))
			
	# Check for overlaps between pairs of transformed wires, and between transformed wires and
	# nontransformed wires. The reason why it is necessary to check for overlap between
	# transformed wires is that a wire may be in 'wires_transformed' even if only one end of it 
	# is transformed.
	for i,(w1,w1pos) in enumerate(wires_transformed):
	
		# Check if w1 would overlap any transformed wires
		for j in xrange(i+1, len(wires_transformed)):
			w2,w2pos = wires_transformed[j]
			if line_segments_overlap(w1pos, w2pos):
				failures.add(w1)
				failures.add(w2)
		
		# Check if w1 would overlap any nontransformed wires
		for (w2,w2pos) in wires_nontransformed:
			if line_segments_overlap(w1pos, w2pos):
				failures.add(w1)
	
	# Check for overlap between transformed components and nontransformed components.
	for (c1,c1bounds) in components_transformed:
		for (c2,c2bounds) in components_nontransformed:
			if rects_intersect(c1bounds, c2bounds):
				failures.add(c1)
	
	# Check for wires ending within a component
	for (w,wpos) in wires_transformed:
		for (c,cbounds) in components_transformed + components_nontransformed:
			for endpos in wpos:
				if point_in_rect(endpos, cbounds):
					failures.add(w)
	
	# Check for wires of length 0 and wires outside of the circuit boundaries
	for (w,wpos) in wires_transformed:
		if wpos[0]==wpos[1]:
			failures.add(w)
		for end in wpos:
			if not point_in_rect(end, ((-1,-1),(size[0]+1,size[1]+1))):
				failures.add(w)
	
	# Check for components outside the circuit boundaries
	for (c,cbounds) in components_transformed:
		if not point_in_rect(cbounds[0], ((-1,-1),(size[0]+1,size[1]+1))):
			failures.add(c)
		if not point_in_rect(add_points(cbounds[0],cbounds[1]), ((-1,-1),(size[0]+1,size[1]+1))):
			failures.add(c)
	
	return failures

def parts_of_objs(objs):
	parts = set()
	for obj in objs:
		if isinstance(obj, Component):
			parts.add(obj)
		else:
			parts.add(obj.start)
			parts.add(obj.end)
	return parts

def objs_of_parts(parts):
	return set([p if isinstance(p,Component) else p.wire for p in parts])

def group_bounds(group, transform=(None,set())):
	# group_bounds returns a rectangle that is the outer boundaries of all of the parts in the
	# given group. If the group is empty, it returns None.
		
	if group:
		def objiter():
			for obj in group:
				if isinstance(obj, WireEnd):
					if obj in transform[1]:
						yield transform_point(transform[0], obj.location)
					else:
						yield obj.location
				else:
					if obj in transform[1]:
						untransformed = obj.bounds()
						ll = untransformed[0]
						ur = add_points(ll, untransformed[1])
						yield rect_with_corners(transform_point(transform[0], ll),
												transform_point(transform[0], ur))
					else:
						yield obj.bounds()
		return rect_bounding_objects(objiter())
	else:
		return None